/*
------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------
*********************************************** 602_Projekt_2_Draht_LCD ************************************************
------------------------------------------------------------------------------------------------------------------------
**************************************************** main.c ** V1.1 ****************************************************
------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------
 Ansteuerung einer 20x4-LCD-Anzeige mittels eines 2-Draht-LCD-Interfaces
 Zusaetzliche Header..: keine
 Module (*.c und *.h).: application, lcd_2wire, num_conversion
 Benoetigte Hardware..: Testboard, Taste S1, 20x4-LCD und 2-Draht-LCD-Interface
 Version..............: 1.1
 Compiler.............: CodeVisionAVR
 Chip.................: ATmega88
 Datum................: Juni 2009
 Autor................: Alwin Lenck (ALE23) basierend auf dem Projekt LCD von Udo Juerss
------------------------------------------------------------------------------------------------------------------------

Dieses Projekt stellt eine Applikation zum Gebrauch des 2-Draht-LCD-Interface dar, das dazu geeignet ist, fuer alle
weiteren Projekte, die eine 20x4-LCD-Anzeige benoetigen, eingesetzt zu werden. Es liegt die letzte Version 1.2 von
September 2008 mit zusaetzlichen Funktionen zu Grunde.

Der Treiber lcd_2wire ist fuer die Nutzung des Controllers HD44780 (oder kompatible) konzipiert. Die Schaltung
ermoeglicht durch die Zwischenschaltung eines Shift-Registers 4094 die Ansteuerung eines LC-Displays im 4-Bit-Mode
lediglich ueber 2 Port-Pins des Mikrocontrollers ATmega88.

ERKLAERUNGEN ZUR SOFTWARE (STEUERUNG UND UEBERTRAGUNGS-MODI)

Zu den Uebertragungs-Modi (char-Variable mode):
    Modus 0 (LCD-Kommando-Uebertragung):
             LOW in das Shift-Register fuer das Daten/Kommando-Signal RS (landet letztlich durch Shift auf Q6)
    Modus 1 (LCD-Daten-Uebertragung):
             HIGH in das Shift-Register fuer das Daten/Kommando-Signal RS (landet letztlich durch Shift auf Q6)
    Modus 2 (LCD-Kommando-Uebertragung fuer die LCD-Initialisierung):
             LOW in das Shift-Register fuer das Daten/Kommando-Signal RS (landet letztlich durch Shift auf Q6)

 Ein HIGH auf Q7 fuer das Enable-Signal E muss vor jeder Uebertragung zum LCD gesetzt sein !

 Zur Takt-Erzeugung (CLK) fuer das Shift-Register werden stets die folgenden 3 Befehle benoetigt:
    PORTB  |= 0x02;         an PB1 => CLK = 1 (steigende Impuls-Flanke, sie loest den Shift aus)
    lcd_delay();            Verzoegerung (Dauer je nach eingestelltem System-Takt)
    PORTB  &= ~0x02;        an PB1 => CLK = 0 (fallende Impuls-Flanke)
 Zur Daten-Bit-Erzeugung (DATA) fuer das Shift-Register werden die folgenden Befehle verwendet:
    PORTB  |= 0x04;         an PB2 => DATA = 1
    lcd_delay();            Verzoegerung (Dauer je nach eingestelltem System-Takt)
    PORTB  &= ~0x04;        an PB2 => DATA = 0
    lcd_delay();            Verzoegerung (Dauer je nach eingestelltem System-Takt)

------------------------------------------------------------------------------------------------------------------------
FUNKTIONEN ZUR INITIALISIERUNG DES LCD UND AUSGABE VON KOMMANDOS UND ASCII-DATEN AN DAS LCD
------------------------------------------------------------------------------------------------------------------------
void lcd_init(void)                              // Initialisierung des LC-Displays                                  1.1
void lcd_delay(void)                             // Verzoegerung in Abhaengigkeit der System-Takt-Frequenz           1.2
void lcd_shift_byte(U08 data,U08 mode)           // Schreibe LED-Kommando oder Daten-Byte in das LC-Display          1.3
void lcd_init_shift_reg(U08 mode)                // Loesche Shift-Register, setze E- und ggf. RS-Bit                 1.4
void lcd_shift_nibble(U08 data)                  // Schiebe 4 Bits (ein Nibble) in das Shift-Register                1.5
void lcd_write_command(U08 command)              // Uebertrage ein vollstaendiges LCD-Kommando zum LCD               1.6
void lcd_write_data(U08 data)                    // Uebertrage ein vollstaendiges Daten-Byte zum LCD                 1.7

------------------------------------------------------------------------------------------------------------------------
FUNKTIONEN ZUM ERZEUGEN VON DISPLAY-KOMMANDOS
------------------------------------------------------------------------------------------------------------------------
void lcd_clear(void)                             // CLEAR DISPLAY:                   Bit0 gesetzt                    2.1
void lcd_home(void)                              // RETURN HOME:                     Bit1 gesetzt                    2.2
void lcd_entry_mode(U08 entry_mode)              // SET ENTRY MODE:                  Bit2 gesetzt                    2.3
void lcd_display_mode(U08 display_mode)          // DISPLAY ON/OFF:                  Bit3 gesetzt                    2.4
void lcd_function_mode(U08 function_mode)        // SET FUNCTION:                    Bit5 gesetzt                    2.5
void lcd_download_char_from_rom(U08 char_index,fpU08 byte_ptr) // SET CGRAM ADDRESS: Bit6 gesetzt                    2.6
void lcd_goto_xy(U08 xpos,U08 ypos)              // Setze Cursor-Koordinaten (x/y)   Bit7 gesetzt                    2.7

------------------------------------------------------------------------------------------------------------------------
FUNKTIONEN ZUM ERZEUGEN VON ASCII-DATEN-STRINGS AUS VERSCHIEDENEN ZEICHEN UND ZEICHEN-GRUPPEN
------------------------------------------------------------------------------------------------------------------------
Verwendete Uebergabe-Variablen:
chr                                              // auszugebendes ASCII-Zeichen (unsigned char)
sp                                               // Pointer auf char-String (Array mit ASCCI-Zeichen)
value                                            // auszugebender Zahlen-Wert (8-Bit- oder 16-Bit-Dual-Zahl)
digits                                           // Anzahl auszugebender Zeichen; ggf. Auffuellen mit
                                                 // vorlaufenden Leerstellen (Spaces ' ')
Verwendete Typen:
U08                                              // unsigned char
pU08                                             // Pointer auf eine char-Variable
fpU08                                            // Pointer auf eine char-Variable im FLASH-Speicher

void lcd_write_char(U08 chr)                     // Schreibe einzelnes ASCII-Zeichen (unsigned char)                 3.1
void lcd_write_char_xy(U08 x,U08 y,U08 chr)      // Schreibe ASCII-Zeichen auf vorbestimmte Position im LCD          3.2
void lcd_write_str(pU08 sp)                      // Schreibe ASCII-String (Array mit ASCCI-Zeichen)                  3.3
void lcd_write_str_xy(U08 x,U08 y,pU08 sp)       // Schreibe ASCII-String auf vorbestimmte Position im LCD           3.4
void lcd_write_flash_str(fpU08 sp)               // Schreibe ASCII-String aus dem Array im FLASH-Speicher            3.5
void lcd_write_flash_str_xy(U08 x,U08 y,fpU08 sp) // Schreibe ASCII-String vom FLASH auf feste Cursor-Position       3.6
void lcd_write_U08(U08 value,U08 digits)         // Schreibe 8-Bit-Dual-Zahl als Dezimal-Zahl                        3.7
void lcd_write_S08(S08 value,U08 digits)         // Schreibe +/- 8-Bit-Dual-Zahl als Dezimal-Zahl                    3.8
void lcd_write_U08_hex(U08 value)                // Schreibe 8-Bit-Dual-Zahl als 2-ziffrige Hexadezimal-Zahl         3.9
void lcd_write_U08_bin(U08 value)                // Schreibe 8-bit-Dual-Zahl in binaerer Schreibweise               3.10
void lcd_write_U16(U16 value,U08 digits)         // Schreibe 16-Bit-Dual-Zahl als Dezimal-Zahl                      3.11
void lcd_write_S16(S16 value,U08 digits)         // Schreibe +/- 16-Bit-Dual-Zahl als Dezimal-Zahl                  3.12
void lcd_write_U16_hex(U16 value)                // Schreibe 16-Bit-Dual-Zahl als 4-ziffrige Hexadezimal-Zahl       3.13
void lcd_write_U32(U32 value,U08 digits)         // Schreibe 32-Bit-Dual-Zahl als Dezimal-Zahl                      3.14
void lcd_write_S32(S32 value,U08 digits)         // Schreibe +/- 32-Bit-Dual-Zahl als Dezimal-Zahl                  3.15
void lcd_write_float(float value,U08 decimals,U08 digits) // Schreibe +/- Gleitkomma-Zahl als Dezimalbruch-Zahl     3.16

------------------------------------------------------------------------------------------------------------------------
FUNKTIONEN FUER DIE KONVERTIERUNG VERSCHIEDENER DATEN-TYPEN
------------------------------------------------------------------------------------------------------------------------
pU08 nc_format(pU08 source_ptr,U08 digits)       // Formatierung des Uebergabe-Strings durch Auffuellen mit Space    4.1
pU08 nc_U08_to_str(U08 value,U08 digits)         // Wandle eine 8-Bit-Dual-Zahl in ein ASCII-String um               4.2
pU08 nc_S08_to_str(S08 signed_value,U08 digits)  // Wandle eine vorzeichenbehaftete 8-Bit-Dual-Zahl nach ASCII um    4.3
pU08 nc_U08_to_hex(U08 value)                    // Wandle eine 8-Bit-Dual-Zahl in eine Hexadezimal-Zahl um          4.4
pU08 nc_U08_to_bin(U08 value)                    // Wandle eine 8-Bit-Dual-Zahl in ein Binary-String um              4.5
pU08 nc_U16_to_str(U16 value,U08 digits)         // Wandle eine 16-Bit-Dual-Zahl in ein ASCII-String um              4.6
pU08 nc_S16_to_str(S16 signed_value,U08 digits)  // Wandle eine vorzeichenbehaftete 16-Bit-Dual-Zahl nach ASCII      4.7
pU08 nc_U16_to_hex(U16 value)                    // Wandle eine 16-Bit-Dual-Zahl in eine Hexadezimal-Zahl um         4.8
pU08 nc_U32_to_str(U32 value,U08 digits)         // Wandle eine 32-Bit-Dual-Zahl in ein ASCII-String um              4.9
pU08 nc_S32_to_str(S32 signed_value,U08 digits)  // Wandle eine vorzeichenbehaftete 32-Bit-Dual-Zahl nach ASCII     4.10
pU08 nc_U32_to_hex(U32 value)                    // Wandle eine 32-Bit-Dual-Zahl in eine Hexadezimal-Zahl um        4.11
pU08 nc_float_to_str(float value,U08 decimals)   // Wandle eine 32-Bit-Gleitkomma-Zahl in ein ASCII-String um       4.12

------------------------------------------------------------------------------------------------------------------------
DAS HAUPTPROGRAMM main.c
------------------------------------------------------------------------------------------------------------------------
Anmerkung zur Benutzung des 2-Draht-LCD's fuer weitere (auch selbst gestrickte) Projekte:
Im Bezug auf die LCD-Ausgabe im Hauptprogramm main.c sind die am Ende im Zeilen-Kommentar mit &&&&& markierten
Anweisungen gegen Anweisungen des neuen Projektes auszutauschen. Lediglich die Aufrufe der Initialisierungs-
Funktionen app_init() und lcd_init() sowie die hier nicht dargestellten Preprozessor-Anweisungen #include XXXXX sind
im Hauptprogramm main.c zu uebernehmen.

Die Dateien der Module application (*.c und *.h), lcd_2wire (*.c und *.h) und num_conversion (*.c und *.h) sowie
die Header-Dateien typedefs.h, iomx8.h, macros.h, switches.h werden verwendet und sind mit dem Hauptprogramm main.c
im zugehoerigen Projekt-Ordner anzulegen. Anpassungen in der application.h sind von Fall zu Fall vorzunehmen.

Besonderheiten und Einstellungen des hier verwendeten Hauptprogramms:                                              &&&&&
Zu Beginn zeigt das LCD folgendes Bild:                                                                            &&&&&
Row 0        0 0000          Zeilen-Nummer 0    dezimale und hexadezimale Zaehlung bei     0 (dezimal) beginnend   &&&&&
Row 1    16384 4000          Zeilen-Nummer 1    dezimale und hexadezimale Zaehlung bei 16384 (dezimal) beginnend   &&&&&
Row 2    32768 8000          Zeilen-Nummer 2    dezimale und hexadezimale Zaehlung bei 32768 (dezimal) beginnend   &&&&&
Row 3    49152 C000          Zeilen-Nummer 3    dezimale und hexadezimale Zaehlung bei 49152 (dezimal) beginnend   &&&&&
                                                                                                                   &&&&&
main-Schleife:                                                                                                     &&&&&
Durch Druecken der geschalteten S1-Taste wird das Hochzaehlen und die Display-Anzeige angehalten.                  &&&&&
                                                                                                                   &&&&&
In der Header-Datei "application.h" koennen die Ports und Pins neu definiert werden.                               &&&&&

Testboard-Beschaltung fuer das AVR-Projekt 2-Draht-LCD:                                                            &&&&&
CLOCK                          PB1 (aenderbar in "application.h")                                                  &&&&&
DATA                           PB2 (aenderbar in "application.h")                                                  &&&&&
                                                                                                                   &&&&&
Tasten:                                                                                                            &&&&&
S1                             PB3 (aenderbar in "application.h")                                                  &&&&&
*/
//----------------------------------------------------------------------------------------------------------------------
// Header-Dateien einfuegen:
//----------------------------------------------------------------------------------------------------------------------
#include "application.h"                         // Initialisierung der Anwendung
#include "lcd_2wire.h"                           // Initialisierung des 2Draht-LCDs
#include "num_conversion.h"                      // Initialisierung der Zeichen-Konvertierung
//----------------------------------------------------------------------------------------------------------------------
// Globale Variablen:
//----------------------------------------------------------------------------------------------------------------------
U08 n;                                           // Zaehler vom Typ unsigned char                                  &&&&&
U16 counters[LCD_ROWS];                          // Array mit 4 Elementen (LCD-Zeilen) vom Typ unsigned int        &&&&&

//----------------------------------------------------------------------------------------------------------------------
// Hauptprogramm main
//----------------------------------------------------------------------------------------------------------------------
void main(void)
{
  // Setzen des System-Taktes (CPU-Clock), Ausschalten nicht benutzter Hardware-Module
  app_init();                                    // Initialisierung der Anwendung

  // Definition der Funktions-Prototypen usw.
  lcd_init();                                    // Initialisierung des LCD-Interfaces

  // Initialisieren der Taste S1                                                                                   &&&&&
  S1_INIT();                                     //                                                                &&&&&

  // Schreibe "Row y" in alle LCD-Zeilen         // y ist die Zeilen-Nummer                                        &&&&&
  for (n = 0; n < LCD_ROWS; n++)                 //                                                                &&&&&
  {                                              //                                                                &&&&&
    lcd_write_flash_str_xy(0,n,"Row ");          // Schreibe "Row " in das LCD, x=0, y=n                           &&&&&
    lcd_write_char(n + 48);                      // Schreibe Zeilen-Nummer als ASCII-Zeichen in das LCD            &&&&&
                                                 // 48 ist der ASCII-Offset fuer Ziffern, d.h.                     &&&&&
  }                                              // 8-bit-Dual-Zahl + 48 ergibt den Code fuer das ASCII-Zeichen    &&&&&
  // Setup fuer Zeilen-Counter mit vorbesetzten aufsteigenden Zahlen                                               &&&&&
  for (n = 0; n < LCD_ROWS; n++)                 //                                                                &&&&&
  {                                              //                                                                &&&&&
    counters[n] = n * 16384;                     // Row 0 beginnt mit counters[0] =      0 = 0x0000                &&&&&
                                                 // Row 1 beginnt mit counters[1] =  16384 = 0x4000                &&&&&
                                                 // Row 2 beginnt mit counters[2] =  32768 = 0x8000                &&&&&
                                                 // Row 3 beginnt mit counters[3] =  49152 = 0xC000                &&&&&
  }                                              //                                                                &&&&&
  // Main loop - Unendliche Schleife                                                                               &&&&&
  while (true)                                   //                                                                &&&&&
  {                                              //                                                                &&&&&
    // So lange wie S1 gedrueckt ist, wird die Zaehlung gestoppt und die Anzeige "eingefroren"                     &&&&&
    S1_WAIT_NOT_PRESSED();                       // So lange S1 gedrueckt ist => TRUE => Schleife ((Modulo))       &&&&&
    // Fortlaufende Zaehlung je Zeile bis 65535 = 0xFFFF                                                           &&&&&
    for (n = 0; n < LCD_ROWS; n++)               // S1 nicht (mehr) gedrueckt => Fortsetzung der Zaehlung          &&&&&
    {                                            //                                                                &&&&&
      // Schreibe den Counter der Zeile rechtsbuendig auf die angegebene Position (Leerzeichen zaehlen mit!)       &&&&&
      lcd_goto_xy(LCD_COLS - 10,n);              // Setze Cursor-Position auf x=10, y=n                            &&&&&
      lcd_write_U16(counters[n],5);              // Schreibe den Counter dezimal mit 5 Zeichen (incl. Leerz.)      &&&&&
      // Schreibe den Counter der Zeile hexadezimal (4 hexadezimale Ziffern) auf die angegebene Position           &&&&&
      lcd_goto_xy(LCD_COLS - 4,n);               // Setze Cursor-Position auf x=16, y=n                            &&&&&
      lcd_write_U16_hex(counters[n]++);          // Schreibe den Counter hexadezimal und erhoehe ihn um 1          &&&&&
    }                                            //                                                                &&&&&
  }                                              //                                                                &&&&&
}
//----------------------------------------------------------------------------------------------------------------------
